
"""

GrabberSnap - Canvas module
SegPhault (Ryan Paul) - 12/07/2008

"""

import gtk, gobject
import goocanvas, cairo

PROXIMITY = 3

def gdk_to_rgba(c, alpha=255):
  return (((c.red / 0x100) << 24) | ((c.green / 0x100) << 16) | ((c.blue / 0x100) << 8) | alpha)

def convert_color(*args):
  if isinstance(args[0], str):
    return gdk_to_rgba(gtk.gdk.color_parse(args[0]),
      args[1] if len(args) > 1 else 255)
  else:
    return gdk_to_rgba(gtk.gdk.Color(args[0] * (65535/255),
        args[1]* (65535/255), args[2]* (65535/255)),
      args[3] if len(args) > 3 else 255)

class DraggableItem(goocanvas.ItemSimple):
  __gtype_name__ = "DraggableItem"
  def __init__(self, *args, **kargs):
    goocanvas.ItemSimple.__init__(self, *args, **kargs)
    self.connect("motion-notify-event", self.on_motion)
    self.connect("button-press-event", self.on_press)
    self.connect("button-release-event", self.on_release)

  def on_motion(self, rect, rect2, e):
    x, y = e.get_coords()
    if rect.get_data("dragging"):
      diffx, diffy = rect.get_data("difference")
      rect.props.x = x - diffx
      rect.props.y = y - diffy
    
    if rect.get_data("dragging"):
      self.get_canvas().window.set_cursor(gtk.gdk.Cursor(gtk.gdk.FLEUR))

  def on_press(self, rect, rect2, e):
    x, y = e.get_coords()
    rect.set_data("difference", (x - rect.props.x, y - rect.props.y))
    rect.set_data("dragging", True)

  def on_release(self, rect, rect2, e):
    rect.set_data("dragging", False)
    self.get_canvas().window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))

class SmartGroup(goocanvas.Group):
  # This class is from goocanvashelper.py: http://is.gd/ayUJ
  __gtype_name__ = 'SmartGroup'

  x = gobject.property(type=int, default=0)
  y = gobject.property(type=int, default=0)
  width = gobject.property(type=int, default=0)
  height = gobject.property(type=int, default=0)

  def __init__(self, *args, **kwargs):
    goocanvas.Group.__init__(self, *args, **kwargs)
    self.children = {}
    self.connect("notify::x", self.move_x_children)
    self.connect("notify::y", self.move_y_children)

  def move_x_children(self, object, prop):
    for child, (x, y) in self.children.items():
      child.props.x = self.x + x

  def move_y_children(self, object, prop):
    for child, (x, y) in self.children.items():
      child.props.y = self.y + y

  def add_child(self, child, pos=(0,0)):
    goocanvas.Group.add_child(self, child)
    self.children[child] = pos

class DraggableText(goocanvas.Text, DraggableItem):
  pass

class Label(SmartGroup, DraggableItem):
  __gtype_name__ = "Label"
  def __init__(self, shadow=False, shadow_color=None, shadow_offset=2, **kargs):
    SmartGroup.__init__(self)
    DraggableItem.__init__(self)
    self.label = goocanvas.Text(**kargs)
    
    if shadow:
      kargs["fill_color_rgba"] = shadow_color
      kargs["x"] = self.label.props.x + shadow_offset
      kargs["y"] = self.label.props.y + shadow_offset
      self.shadow = goocanvas.Text(**kargs)
      self.add_child(self.shadow, (self.shadow.props.x, self.shadow.props.y))
    
    self.add_child(self.label, (self.label.props.x, self.label.props.y))

class ResizableRect(goocanvas.Rect):
  __gtype_name__ = "ResizableRect"
  def __init__(self, *args, **kargs):
    goocanvas.Rect.__init__(self, *args, **kargs)
    self.connect("motion-notify-event", self.on_motion)
    self.connect("leave-notify-event", self.on_leave)
    self.connect("button-press-event", self.on_press)
    self.connect("button-release-event", self.on_release)
    self.constrain = kargs["constrain"] if "constrain" in kargs else None

  def get_rect_edge_proximity(self, r, x, y, prox = PROXIMITY):
    return {
      "top": abs(y - r.props.y) < prox,
      "left": abs(x - r.props.x) < prox,
      "right": abs(x - (r.props.x + r.props.width)) < prox,
      "bottom": abs(y - (r.props.y + r.props.height)) < prox}
   
  def get_prox_cursor(self, p):
    if p["top"] and p["left"]: return gtk.gdk.TOP_LEFT_CORNER
    elif p["top"] and p["right"]: return gtk.gdk.TOP_RIGHT_CORNER
    elif p["bottom"] and p["left"]: return gtk.gdk.BOTTOM_LEFT_CORNER
    elif p["bottom"] and p["right"]: return gtk.gdk.BOTTOM_RIGHT_CORNER
    elif p["top"]: return gtk.gdk.TOP_SIDE
    elif p["bottom"]: return gtk.gdk.BOTTOM_SIDE
    elif p["left"]: return gtk.gdk.LEFT_SIDE
    elif p["right"]: return gtk.gdk.RIGHT_SIDE
    else: return gtk.gdk.ARROW

  def resize_rect(self, rect, prox, x, y):
    cx, cy, cw, ch = self.get_canvas().get_bounds()

    if prox["bottom"]:
      if y > ch and self.constrain: rect.props.height = ch - rect.props.y
      else: rect.props.height = y - rect.props.y
    if prox["right"]:
      if x > cw and self.constrain: rect.props.width = cw - rect.props.x
      else: rect.props.width = x - rect.props.x
    if prox["left"]:
      if x < 0 and self.constrain: rect.props.x = 1
      else:
        if x < rect.props.x: rect.props.width += abs(x - rect.props.x)
        else: rect.props.width -= abs(x - rect.props.x)
        rect.props.x = x
    if prox["top"]:
      if y < 0 and self.constrain: rect.props.y = 1
      else:
        if y < rect.props.y: rect.props.height += abs(y - rect.props.y)
        else: rect.props.height -= abs(y - rect.props.y)
        rect.props.y = y

  def on_motion(self, rect, rect2, e):
    x, y = e.get_coords()
    cx, cy, cw, ch = self.get_canvas().get_bounds()

    if rect.get_data("dragging"):
      if True in rect.get_data("direction").values():
        self.resize_rect(rect, rect.get_data("direction"), x, y)
      else:
        diffx, diffy = rect.get_data("difference")
        if self.constrain:
          if x - diffx < 0: rect.props.x = 1
          elif (x - diffx) + rect.props.width > cw:
            rect.props.x = cw - rect.props.width
          else: rect.props.x = x - diffx
          
          if y - diffy < 0: rect.props.y = 1
          elif (y - diffy) + rect.props.height > ch:
            rect.props.y = ch - rect.props.height
          else: rect.props.y = y - diffy
        else:
          rect.props.x = x - diffx
          rect.props.y = y - diffy
    
    prox = self.get_rect_edge_proximity(rect, *e.get_coords())
    ct = self.get_prox_cursor(prox)
    if rect.get_data("dragging") and ct == gtk.gdk.ARROW: ct = gtk.gdk.FLEUR
    self.get_canvas().window.set_cursor(gtk.gdk.Cursor(ct))

  def on_leave(self, rect, rect2, e):
    if not rect.get_data("dragging"):
      self.get_canvas().window.set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))

  def on_press(self, rect, rect2, e):
    x, y = e.get_coords()
    prox = self.get_rect_edge_proximity(rect, x, y)
    rect.set_data("difference", (x - rect.props.x, y - rect.props.y))
    rect.set_data("dragging", True)
    rect.set_data("direction", prox)

  def on_release(self, rect, rect2, e):
    rect.set_data("dragging", False)
